# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Linear algebra utility functions
"""
# pylint: disable=W0611


from cr.nimble._src.util import (
    platform,
    is_cpu,
    is_gpu,
    is_tpu,
    KEY0,
    KEYS,
    promote_arg_dtypes,
    canonicalize_dtype,
    promote_to_complex,
    promote_to_real,
    integer_types,
    integer_ranges,
    dtype_ranges,
    nbytes_live_buffers,
)

from cr.nimble._src.discrete.number import (
    next_pow_of_2,
    is_integer,
    is_positive_integer,
    is_negative_integer,
    is_odd,
    is_even,
    is_odd_natural,
    is_even_natural,
    is_power_of_2,
    is_perfect_square,
    integer_factors_close_to_sqr_root
)

from cr.nimble._src.array import (
    hermitian,
    check_shapes_are_equal
)

from cr.nimble._src.vector import (
    is_scalar,
    is_vec,
    is_line_vec,
    is_row_vec,
    is_col_vec,
    to_row_vec,
    to_col_vec,
    vec_unit,
    vec_unit_jit,
    vec_shift_right,
    vec_rotate_right,
    vec_shift_left,
    vec_rotate_left,
    vec_shift_right_n,
    vec_rotate_right_n,
    vec_shift_left_n, 
    vec_rotate_left_n,
    vec_safe_divide_by_scalar,   
    vec_repeat_at_end,
    vec_repeat_at_end_jit,
    vec_repeat_at_start,
    vec_repeat_at_start_jit,
    vec_centered,
    vec_centered_jit,
    vec_to_windows,
    vec_to_windows_jit,
    vec_mag_desc,
    vec_to_pmf,
    vec_to_cmf,
    cmf_find_quantile_index,
    num_largest_coeffs_for_energy_percent
)

from cr.nimble._src.matrix import (
    AH_v,
    mat_transpose,
    mat_hermitian,
    is_matrix,
    is_square,
    is_symmetric,
    is_hermitian,
    is_positive_definite,
    has_orthogonal_columns,
    has_orthogonal_rows,
    has_unitary_columns,
    has_unitary_rows,
    off_diagonal_elements,
    off_diagonal_min,
    off_diagonal_max,
    off_diagonal_mean,
    set_diagonal,
    add_to_diagonal,
    abs_max_idx_cw,
    abs_max_idx_rw,
    diag_premultiply,
    diag_postmultiply
)



from cr.nimble._src.norm import (
    norm_l1,
    sqr_norm_l2,
    norm_l2,
    norm_linf,

    normalize_l1,
    normalize_l2,
    normalize_linf,

    norms_l1_cw,
    norms_l1_rw,
    norms_l2_cw,
    norms_l2_rw,
    norms_linf_cw,
    norms_linf_rw,
    sqr_norms_l2_cw,
    sqr_norms_l2_rw,


    normalize_l1_cw,
    normalize_l1_rw,
    normalize_l2_cw,
    normalize_l2_rw,
)


from cr.nimble._src.distance import (
    pairwise_sqr_l2_distances_rw,
    pairwise_sqr_l2_distances_cw,
    pairwise_l2_distances_rw,
    pairwise_l2_distances_cw,
    pdist_sqr_l2_rw,
    pdist_sqr_l2_cw,
    pdist_l2_rw,
    pdist_l2_cw,
    # Manhattan distances
    pairwise_l1_distances_rw,
    pairwise_l1_distances_cw,
    pdist_l1_rw,
    pdist_l1_cw,

    # Chebychev distance
    pairwise_linf_distances_rw,
    pairwise_linf_distances_cw,
    pdist_linf_rw,
    pdist_linf_cw
)


from cr.nimble._src.metrics import (
    mean_squared,
    mean_squared_error,
    root_mean_squared,
    root_mse,
    normalization_factor,
    normalized_root_mse,
    peak_signal_noise_ratio,
    signal_noise_ratio
)

from cr.nimble._src.ndarray import (
    arr_largest_index,
    arr_l1norm,
    arr_l2norm,
    arr_l2norm_sqr,
    arr_vdot,
    arr_rdot,
    arr_rnorm_sqr,
    arr_rnorm,
    arr2vec,
    log_pos,
)


from cr.nimble._src.linear import (
    point2d,
    vec2d,
    rotate2d_cw,
    rotate2d_ccw,
    reflect2d,
)

from cr.nimble._src.triangular import (
    solve_Lx_b,
    solve_LTx_b,
    solve_Ux_b,
    solve_UTx_b,
    solve_spd_chol
)

from cr.nimble._src.householder import (
    householder_vec,
    householder_matrix,
    householder_premultiply,
    householder_postmultiply,
    householder_ffm_jth_v_beta,
    householder_ffm_premultiply,
    householder_ffm_backward_accum,
    householder_ffm_to_wy,
    householder_qr_packed,
    householder_split_qf_r,
    householder_qr,
)

from cr.nimble._src.chol import (
    cholesky_update_on_add_column,
    cholesky_build_factor
)

# These functions are not JIT ready
from cr.nimble._src.householder import (
    householder_vec_
)

from cr.nimble._src.svd_utils import (
    orth,
    orth_jit,
    row_space,
    row_space_jit,
    null_space,
    null_space_jit,
    left_null_space,
    left_null_space_jit,
    effective_rank,
    effective_rank_jit,
    effective_rank_from_svd,
    singular_values
)

from cr.nimble._src.dls import (
    mult_with_submatrix,
    solve_on_submatrix
)


from cr.nimble._src.standard_matrices import (
    gaussian_mtx,
    pascal,
    pascal_jit
)


########################################
# Basic signal processing stuff
########################################

from cr.nimble._src.signal import (
    nonzero_values,
    nonzero_indices,
    support,
    randomize_rows,
    randomize_cols,
    largest_indices,
    hard_threshold,
    hard_threshold_sorted,
    sparse_approximation,
    build_signal_from_indices_and_values,
    hard_threshold_by,
    largest_indices_by,
    dynamic_range,
    nonzero_dynamic_range,

    # row wise
    largest_indices_rw,
    take_along_rows,
    sparse_approximation_rw,

    # column wise
    largest_indices_cw,
    take_along_cols,
    sparse_approximation_cw,
)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple
import math

from jax import lax, jit, vmap, random
import jax.numpy as jnp
from jax.numpy.linalg import norm

FUDGE = 1.01
M2 = 3/2
N2 = 3/2

class LanBDOptions(NamedTuple):
    """Options for Lanczos bidiagonalization with partial reorthogonalization algorithm"""
    delta : float = 0.
    "desired level of orthogonality"
    eta : float = 0.
    "desired level of orthogonality after reorthogonalization"
    gamma : float = 0.
    "Tolerance for iterate Gram-Schmidt"
    cgs : bool = False
    "Flag for Classic / Modified Gram Schmidt"
    elr : int = 2
    "Flag for extended local reorthogonalization"
    eps : float = jnp.finfo(float).eps
    "eps value"

def lanbpro_options_init(k : int, eps: float =jnp.finfo(float).eps):
    """Initializes options"""
    delta = jnp.sqrt(eps/k)
    eta = (eps ** 0.75) / jnp.sqrt(k)
    gamma = 1 / math.sqrt(2)
    return LanBDOptions(delta=delta, eta=eta, gamma=gamma, eps=eps)

class LanBProState(NamedTuple):
    """State of Lanczos bidiagonalization with partial reorthogonalization algorithm

    At the end of j-iterations
    - alpha has j entries
    - U has j columns
    - V has j columns
    - beta has j+1 entries
    """
    p : jnp.ndarray
    "Starting vector for next left singular vector"
    U : jnp.ndarray
    "Left singular vectors"
    V : jnp.ndarray
    "Right singular vectors"
    alpha: jnp.ndarray
    "Diagoanal elements of B"
    beta: jnp.ndarray
    "Subdiagonal elements of B"
    mu: jnp.ndarray
    "Tracker for U inner products"
    nu: jnp.ndarray
    "Tracker for V inner products"
    mumax: jnp.ndarray
    "Maximum values of mu for each iteration"
    numax: jnp.ndarray
    "Maximum values of nu for each iteration"
    indices: jnp.ndarray
    "Set of indices to orthogonalize against"
    anorms : jnp.ndarray
    "Tracking the a norms over multiple iterations" 
    anorm : float = -1.
    "Running estimate of the norm of A" 
    nreorthu: int = 0
    "Number of reorthogonalizations on U"
    nreorthv: int = 0
    "Number of reorthogonalizations on V"
    npu: int = 0
    "Number of U inner products"
    npv: int = 0
    "Number of V inner products"
    nrenewu: int = 0
    "Number of times u vectors were renewed"
    nrenewv: int = 0
    "Number of times v vectors were renewed"
    b_force_reorth: bool = False
    "Indicates if reorthogonalization is forced in next iter"
    b_fro: bool = False
    "flag to indicate if full reorthogonalization is to be done"
    b_est_anorm: bool = True
    "Flag which indicates whether norm of A is to be estimated"
    iterations: int = 0
    "iteration number for j-th Lanczos vector"

    def __str__(self):
        s = []
        j = self.iterations
        # s.append(f'p= {self.p}')
        # s.append(f'U= {self.U}')
        # s.append(f'V= {self.V}')
        s.append(f'alpha= {self.alpha[:j]}')
        s.append(f'beta= {self.beta[:j+1]}')
        s.append(f'mu= {self.mu[:j]}')
        s.append(f'nu= {self.nu[:j]}')
        s.append(f'mumax= {self.mumax[:j]}')
        s.append(f'numax= {self.numax[:j]}')
        # s.append(f'anorm= {self.anorm}')
        s.append(f'b_fro= {self.b_fro}, b_force_reorth={self.b_force_reorth}')
        ind = jnp.where(self.indices)[0]
        s.append(f'indices= {ind}')
        s.append(f'iterations= {self.iterations}')
        return '\n'.join(s)

def lanbpro_random_start(key, A):
    """Creates a random starting vector for the algorithm"""
    m, n = A.shape
    r = random.uniform(key, (m,)) - 0.5
    return r

def update_mu(mu, mumax, nu, j, alpha, beta, anorm):
    """Updates mu

    At the beginning of this call U[j+1] is being computed.

    * p has been computed
    * alpha[j] has been computed
    * j+1 U columns are already computed.
    * j+1 V columns are already computed.
    * beta[j+1] has been computed.
    * Estimates of inner product of U[j+1] with U_0 till U_j is being updated.s

    For U_1, 1 entry will be updated.
    For U_2, 2 entries will be updated.

    Essentially mu[0 till j] will be updated.
    mu[j+1] will always be set to 1.

    The mu update is broken down into four parts:

    * update mu[0]
    * update mu[0:j]
    * update mu[j]
    * update mu[j+1]
    """
    eps = jnp.finfo(float).eps
    bjp1 = beta[j+1]
    binv = 1/bjp1
    eps1 = 100*eps/2
    a0 = alpha[0]
    b0 = beta[0]
    aj = alpha[j]
    bj = beta[j]
    k = len(alpha)
    # mu[0] calculation
    tmp = a0*nu[0] - aj*mu[0]
    x = jnp.sqrt(aj**2+bjp1**2)
    y = jnp.sqrt(a0**2+b0**2)
    T = eps1*(x + y + anorm)
    tmp = binv * (tmp + jnp.sign(tmp)*T)
    mu = mu.at[0].set(tmp)
    # sub function to update the middle part of mu
    def update_middle_part(mu):
        idx = jnp.arange(k)
        mask = jnp.logical_and(idx > 0, idx < j)
        # alpha between 1 to j-1
        alpha0 = jnp.where(mask, alpha, 0)
        # beta between 1 to j-1
        beta0 = beta[:k]
        beta0 = jnp.where(mask, beta0, 0)
        # nu between 1 to j-1
        nu0 = jnp.where(mask, nu, 0)
        # nu between 0 to j
        nu1 = jnp.roll(nu, 1)
        nu1 = jnp.where(mask, nu1, 0)
        # mu between 1 to j-1
        mu0 = jnp.where(mask, mu, 0)
        tmp = alpha0*nu0 + beta0*nu1 - aj*mu0
        y = jnp.sqrt(alpha0**2+beta0**2)
        T = eps1*(x + y + anorm)
        tmp = binv*(tmp + jnp.sign(tmp)*T)
        # update mu on the range 1 to j-1
        mu = jnp.where(mask, tmp, mu)
        return mu
    # update the middle part only if j > 1
    mu = lax.cond(j > 1, update_middle_part, lambda mu: mu, mu)
    # update mu[j]
    y = jnp.sqrt(aj**2+bj**2)
    T = eps1*(x + y + anorm)
    tmp = bj*nu[j-1]
    mu = mu.at[j].set(binv * (tmp + jnp.sign(tmp)*T))
    # mumax update
    mumax = mumax.at[j].set(jnp.max(jnp.abs(mu) ))
    return mu, mumax

def update_nu(nu, numax, mu, j, alpha, beta, anorm):
    """Updates nu

    At the beginning of this call V_j is being computed.
 
    * r has been computed
    * alpha[j] has been computed
    * j+1 U columns are already computed.
    * j V columns are already computed.
    * beta[j] has been computed.
    * Estimates of inner product of V[j] with V_0 till V_{j-1} is being updated.
    * Thus, total j entries are under consideration for computing the max value.
    * nu[j] = 1 always.
    """
    eps = jnp.finfo(float).eps
    ainv = 1/alpha[j]
    eps1 = 100*eps/2
    aj = alpha[j]
    bj = beta[j]
    x = jnp.sqrt(aj**2 + bj**2)
    k = len(alpha)
    idx = jnp.arange(k)
    j_mask = idx < j
    # alpha from 0 to j-1
    alpha =  jnp.where(j_mask, alpha, 0)
    # beta from 1 to j
    beta = beta[1:]
    beta = jnp.where(j_mask, beta, 0)
    # mu from 0 to j-1
    mu0 = jnp.where(j_mask, mu, 0)
    # nu from 0 to j-1
    nu0 = jnp.where(j_mask, nu, 0)
    # mu from 1 to j
    mu1 = jnp.roll(mu, -1)
    mu1 = jnp.where(j_mask, mu1, 0)
    # compute T
    y = jnp.sqrt(alpha**2 + beta**2)
    T = eps1*(x + y + anorm)
    # update nu
    nu0 = beta*mu1 + alpha*mu0 - bj*nu0
    nu0 =  ainv*(nu0 + jnp.sign(nu0)*T)
    numax = numax.at[j].set(jnp.max(jnp.abs(nu0) ))
    return nu0, numax


def abs_max_boolean_index(mu):
    """Returns a boolean index array which is True only at the place of 
    absolute maximum
    """
    return jnp.zeros(len(mu), dtype=bool).at[jnp.argmax(jnp.abs(mu))].set(True)


def update_r(mu, indices, i, eta):
    """Start from i, go back and set all consecutive entries to True which are above eta.
    """
    state = indices, i-1
    def cond(state):
        indices, j = state
        c = jnp.logical_and(j>=0, mu[j] >= eta)
        c = jnp.logical_and(c, indices[j] == False)
        return c

    def body(state):
        indices, j = state
        indices = indices.at[j].set(True)
        return indices, j-1

    return lax.while_loop(cond, body, state)[0]
    # while cond(state):
    #     state = body(state)
    # return state[0]

def update_s(mu, indices, i, eta):
    """Start from i, go forward and set all consecutive entries to True which are above eta.
    """
    n = len(mu)
    state = indices, i+1
    def cond(state):
        indices, j = state
        c = jnp.logical_and(j<n, mu[j] >= eta)
        c = jnp.logical_and(c, indices[j] == False)
        return c

    def body(state):
        indices, j = state
        indices = indices.at[j].set(True)
        return indices, j+1

    return lax.while_loop(cond, body, state)[0]
    # while cond(state):
    #     state = body(state)
    # return state[0]

def extend_around(mu, indices, i, eta):
    indices = update_r(mu, indices, i, eta)
    indices = update_s(mu, indices, i, eta)
    return indices

def compute_ind(mu, delta, eta):
    """Identifies indices for reorthogonalization
    """
    k = len(mu)
    mu = jnp.abs(mu)
    indices = jnp.zeros(k, dtype=bool)
    indices = mu >= delta
    # number of indices to orthogonalize against
    n = jnp.sum(indices)
    indices = lax.cond(n > 0, lambda _ : indices, lambda _ : abs_max_boolean_index(mu), None)
    # make a copy
    indices2 = indices
    # loop for extending reorthogonalization around already identified indices
    def extend_rs_iter(i, indices):
        return lax.cond(indices2[i], lambda indices :  extend_around(mu, indices, i, eta), 
            lambda indices: indices, indices)
    indices = lax.fori_loop(0, k, extend_rs_iter, indices)
    return indices


def bpro_norm_estimate(alpha, beta):
    """Estimates the norm based on a 6x5 matrix
    """
    k = 5
    if len(alpha) < k:
        alpha = jnp.pad(alpha, (0, k - len(alpha)))
    if len(beta) <= k:
        beta = jnp.pad(beta, (0, k + 1 - len(beta)))
    # prepare the k+1 x k bidiagonal matrix
    B = jnp.zeros((k+1, k))
    # diagonal indices for k alpha entries
    indices = jnp.diag_indices(k)
    B = B.at[indices].set(alpha[:k])
    # subdiagonal indices for k beta entries (from second row)
    rows, cols = indices
    rows = rows + 1
    B = B.at[(rows, cols)].set(beta[1:k+1])
    result = norm(B, 2)
    return result

def do_elr(v_prev, v, v_norm, gamma):
    """
    Extended local reorthogonalization
    """
    def init():
        t = jnp.vdot(v_prev, v)
        v2 = v - t * v_prev
        v2_norm = norm(v2)
        return v2, v2_norm, v_norm, t

    def body(state):
        v, old_norm, older_norm, proj = state
        t = jnp.vdot(v_prev, v)
        v = v - t * v_prev
        v_norm = norm(v)
        return v, v_norm, old_norm, proj + t

    def cond(state):
        v, v_norm, old_norm, proj = state
        return v_norm < gamma * old_norm

    # state = init()
    # while cond(state):
    #     state = body(state)
    state = lax.while_loop(cond, body, init())
    v, v_norm, old_norm, proj = state
    return v, v_norm, proj



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import jit, lax
import jax.numpy as jnp

from .util import promote_arg_dtypes

def AH_v(A, v):
    r"""Returns :math:`A^H v` for a given matrix A and a vector v

    Args:
        A (jax.numpy.ndarray): A matrix
        v (jax.numpy.ndarray): A vector

    Returns:
        (jax.numpy.ndarray): A vector: :math:`A^H v`

    This is definitely faster on large matrices
    """
    return jnp.conjugate((jnp.conjugate(v.T) @ A).T)


def mat_transpose(x):
    """Returns the transpose of an array of matrices

    Args:
        x (jax.numpy.ndarray): An nd-array (2 or more dimensions)

    Returns:
        (jax.numpy.ndarray): Array with last two dimensions transposed
    """
    return jnp.swapaxes(x, -1, -2)


@jit
def mat_hermitian(a):
    """Returns the conjugate transpose of an array of matrices

    Args:
        A (jax.numpy.ndarray): A JAX array (2 or more dimensions)

    Returns:
        jax.numpy.ndarray: Conjugate transpose of the array
    """
    return jnp.conjugate(jnp.swapaxes(a, -1, -2))

@jit
def is_matrix(A):
    """Checks if an array is a matrix

    Args:
        A (jax.numpy.ndarray): A JAX array


    Returns:
        bool: True if the array is a matrix, False otherwise.
    """
    return A.ndim == 2

@jit
def is_square(A):
    """Checks if an array is a square matrix

    Args:
        A (jax.numpy.ndarray): A JAX array


    Returns:
        bool: True if the array is a square matrix, False otherwise.
    """
    shape = A.shape
    return A.ndim == 2 and shape[0] == shape[1]

@jit
def is_symmetric(A):
    """Checks if an array is a symmetric matrix

    Args:
        A (jax.numpy.ndarray): A JAX array


    Returns:
        bool: True if the array is a symmetric matrix, False otherwise.
    """
    if A.ndim != 2: 
        return False
    return jnp.array_equal(A, A.T)

@jit
def is_hermitian(A):
    """Checks if an array is a Hermitian matrix

    Args:
        A (jax.numpy.ndarray): A JAX array


    Returns:
        bool: True if the array is a Hermitian matrix, False otherwise.
    """
    shape = A.shape
    if A.ndim != 2: 
        return False
    if shape[0] != shape[1]:
        return False
    return jnp.allclose(A, mat_hermitian(A), atol=1e-6)

def is_positive_definite(A):
    """Checks if an array is a symmetric positive definite matrix

    Args:
        A (jax.numpy.ndarray): A JAX array


    Returns:
        bool: True if the array is a symmetric positive definite matrix, False otherwise.

    Symmetric positive definite matrices have real and positive eigen values.
    This function checks if all the eigen values are positive. 
    """
    if A.ndim != 2: 
        return False
    A = promote_arg_dtypes(A)
    is_sym = jnp.array_equal(A, A.T)
    # check for eigen values only if we know that the matrix is symmetric
    is_pd = lax.cond(is_sym, 
        lambda _ : jnp.all(jnp.real(jnp.linalg.eigvals(A)) > 0), 
        lambda _ : False,
        None) 
    return is_pd


@jit
def has_orthogonal_columns(A, atol=1e-6):
    """Checks if a matrix has orthogonal columns

    Args:
        A (jax.numpy.ndarray): A JAX real 2D array


    Returns:
        bool: True if the matrix has orthogonal columns, False otherwise.
    """
    G = A.T @ A
    m = G.shape[0]
    I = jnp.eye(m)
    return jnp.allclose(G, I, atol=m*m*atol)


@jit
def has_orthogonal_rows(A, atol=1e-6):
    """Checks if a matrix has orthogonal rows

    Args:
        A (jax.numpy.ndarray): A JAX real 2D array


    Returns:
        bool: True if the matrix has orthogonal rows, False otherwise.
    """
    G = A @ A.T
    m = G.shape[0]
    I = jnp.eye(m)
    return jnp.allclose(G, I, atol=m*m*atol)

@jit
def has_unitary_columns(A):
    """Checks if a matrix has unitary columns

    Args:
        A (jax.numpy.ndarray): A JAX real or complex 2D array


    Returns:
        bool: True if the matrix has unitary columns, False otherwise.
    """
    G = mat_hermitian(A) @ A
    m = G.shape[0]
    I = jnp.eye(m)
    return jnp.allclose(G, I, atol=m*1e-6)

@jit
def has_unitary_rows(A):
    """Checks if a matrix has unitary rows

    Args:
        A (jax.numpy.ndarray): A JAX real or complex 2D array


    Returns:
        bool: True if the matrix has unitary rows, False otherwise.
    """
    G = A @ mat_hermitian(A)
    m = G.shape[0]
    I = jnp.eye(m)
    return jnp.allclose(G, I, atol=m*1e-6)


def off_diagonal_elements(A):
    """Returns the off diagonal elements of a matrix A

    Args:
        A (jax.numpy.ndarray): A real 2D matrix

    Returns:
        (jax.numpy.ndarray): A vector of off-diagonal elements in A
    """
    mask = ~jnp.eye(*A.shape, dtype=bool)
    return A[mask]

def off_diagonal_min(A):
    """Returns the minimum of the off diagonal elements

    Args:
        A (jax.numpy.ndarray): A real 2D matrix

    Returns:
        (float): The smallest off-diagonal element in A
    """
    off_diagonal_entries = off_diagonal_elements(A)
    return jnp.min(off_diagonal_entries)

def off_diagonal_max(A):
    """Returns the maximum of the off diagonal elements

    Args:
        A (jax.numpy.ndarray): A real 2D matrix

    Returns:
        (float): The largest off-diagonal element in A
    """
    off_diagonal_entries = off_diagonal_elements(A)
    return jnp.max(off_diagonal_entries)

def off_diagonal_mean(A):
    """Returns the maximum of the off diagonal elements

    Args:
        A (jax.numpy.ndarray): A real 2D matrix

    Returns:
        (float): The mean of all off-diagonal elements in A
    """
    off_diagonal_entries = off_diagonal_elements(A)
    return jnp.mean(off_diagonal_entries)

@jit
def set_diagonal(A, value):
    """Sets the diagonal elements to a specific value

    Args:
        A (jax.numpy.ndarray): A 2D matrix
        value (float) : A value to be added to the diagonal elements

    Returns:
        (jax.numpy.ndarray): Matrix with updated diagonal
    """
    indices = jnp.diag_indices(A.shape[0])
    return A.at[indices].set(value)

@jit
def add_to_diagonal(A, value):
    """Add a specific value to the diagonal elements 

    Args:
        A (jax.numpy.ndarray): A 2D matrix
        value (float) : A value to be added to the diagonal elements

    Returns:
        (jax.numpy.ndarray): Matrix with updated diagonal
    """
    indices = jnp.diag_indices(A.shape[0])
    return A.at[indices].add(value)

@jit
def abs_max_idx_cw(A):
    """Returns the index of entry with highest magnitude in each column
    """
    return jnp.argmax(jnp.abs(A), axis=0)

@jit
def abs_max_idx_rw(A):
    """Returns the index of entry with highest magnitude in each row
    """
    return jnp.argmax(jnp.abs(A), axis=1)



@jit
def diag_premultiply(d, A):
    """Compute D @ A where D is a diagonal matrix with entries from vector d
    """
    return jnp.multiply(d[:, None], A)

@jit
def diag_postmultiply(A, d):
    """Compute A @ D where D is a diagonal matrix with entries from vector d
    """
    return jnp.multiply(A, d)



"""A setuptools based setup module.
"""
import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

# Always prefer setuptools over distutils
from setuptools import setup, find_namespace_packages
# To use a consistent encoding
from codecs import open
from os import path

exec(open('src/cr/nimble/version.py').read())
here = path.abspath(path.dirname(__file__))

def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def _parse_requirements(filename):
  with open(path.join(here, 'requirements', filename)) as f:
    return [
        line.rstrip()
        for line in f
        if not (line.isspace() or line.startswith('#'))
    ]

setup(
    name='cr-nimble',

    version=__version__,

    description='Iterative algorithms for numerical linear algebra with JAX',
    long_description=long_description,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url='https://carnotresearch.github.io/cr-nimble',
    download_url=f"https://github.com/carnotresearch/cr-nimble/archive/v{__version__}.tar.gz",

    # Author details
    author='CR-Suite Development Team',
    author_email='contact@carnotresearch.com',

    # Choose your license
    license='Apache 2.0: http://www.apache.org/licenses/LICENSE-2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Video',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Image Recognition',
        # License
        'License :: OSI Approved :: Apache Software License',
        # OS Support
        'Operating System :: Unix',
        'Operating System :: POSIX',
        # 'Operating System :: Microsoft :: Windows',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    project_urls={
        'Issue Tracker': "https://github.com/carnotresearch/cr-nimble/issues"
    },
    # What does your project relate to?
    keywords='Linear Algebra',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_namespace_packages('src', include=['cr.*']),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    python_requires=">=3.7",
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=_parse_requirements('requirements.txt'),
    tests_require=_parse_requirements('requirements-tests.txt'),
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': [ ],
        'test': _parse_requirements('requirements-tests.txt'),
    },
    include_package_data=True,
    zip_safe=False,

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
        ],
    },
)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Solve a symmatric positive definite system :math:`A x = b`
using Jacobi iterations
"""

from typing import NamedTuple

from jax import lax, jit
import jax.numpy as jnp


class State(NamedTuple):
    x: jnp.ndarray
    """The solution"""
    e_norm_sqr: jnp.ndarray
    """The norm squared of the error between successive x approximations"""
    iterations: int
    """The number of iterations it took to complete"""

def jacobi_solve(A, b, max_iters=None, res_norm_rtol=1e-4):
    """Solves the problem :math:`Ax  = b` for a symmetric positive definite :math:`A` via conjugate gradients iterations
    """
    # Boyd Conjugate Gradients slide 22
    m, n = A.shape

    # Get the diagonal elements
    D = jnp.diag(A)
    # Invert them
    Dinv = 1 / D
    # Set the diagonal elements to 0 to get E
    E = A.at[jnp.diag_indices(m)].set(0)
    # Compute B = D^{-1} E
    B = -jnp.multiply(Dinv[:, None], E)
    # Compute z D^{-1} b 
    z = jnp.multiply(Dinv, b)

    b_norm_sqr = b.T @ b

    max_e_norm_sqr = b_norm_sqr * (res_norm_rtol ** 2)
    if max_iters is None:
        max_iters = 500

    def init():
        x = z
        e_norm_sqr = x.T @ x
        return State(x=x,
            e_norm_sqr=e_norm_sqr,
            iterations=1)

    def iteration(state):
        # update the solution x
        x = B @ state.x + z
        # update the residual r
        r = x - state.x
        e_norm_sqr = r.T @ r
        # update state
        return State(x=x,
            e_norm_sqr=e_norm_sqr,
            iterations=state.iterations+1)

    def cond(state):
        # limit on residual norm 
        a = state.e_norm_sqr > max_e_norm_sqr
        # limit on number of iterations
        b = state.iterations < max_iters
        c = jnp.logical_and(a, b)
        return c

    state = lax.while_loop(cond, iteration, init())
    return state

jacobi_solve_jit  = jit(jacobi_solve,
    static_argnames=("max_iters", "res_norm_rtol"))



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

def homogenize_vec(x, value=1):
    assert x.ndim == 1
    return jnp.hstack((x, value))

def homogenize_cols(X, value=1):
    assert X.ndim == 2
    n = X.shape[-1]
    o = value * jnp.ones(n)
    return jnp.vstack((X, o))


def homogenize(X, value=1):
    if X.ndim == 1:
        return homogenize_vec(X, value)
    return homogenize_cols(X, value)


# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

References

- https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html
- https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.linalg.norm.html

"""

import jax.numpy as jnp

EPS = jnp.finfo(jnp.float32).eps

norm = jnp.linalg.norm

from .util import promote_arg_dtypes

def norm_l1(x):
    """
    Computes the l_1 norm of a vector
    """
    return jnp.sum(jnp.abs(x))

def sqr_norm_l2(x):
    """
    Computes the squared l_2 norm of a vector
    """
    return x.T @ x

def norm_l2(x):
    """
    Computes the l_2 norm of a vector
    """
    return jnp.sqrt(x.T @ x)

def norm_linf(x):
    """
    Computes the l_inf norm of a vector
    """
    return jnp.max(jnp.abs(x))

def norms_l1_cw(X):
    """
    Computes the l_1 norm of each column of a matrix
    """
    return norm(X, ord=1, axis=0)

def norms_l1_rw(X):
    """
    Computes the l_1 norm of each row of a matrix
    """
    return norm(X, ord=1, axis=1)

def norms_l2_cw(X):
    """
    Computes the l_2 norm of each column of a matrix
    """
    return norm(X, ord=2, axis=0, keepdims=False)

def norms_l2_rw(X):
    """
    Computes the l_2 norm of each row of a matrix
    """
    return norm(X, ord=2, axis=1, keepdims=False)


def norms_linf_cw(X):
    """
    Computes the l_inf norm of each column of a matrix
    """
    return norm(X, ord=jnp.inf, axis=0)

def norms_linf_rw(X):
    """
    Computes the l_inf norm of each row of a matrix
    """
    return norm(X, ord=jnp.inf, axis=1)


def sqr_norms_l2_cw(X):
    """
    Computes the squared l_2 norm of each column of a matrix
    """
    return jnp.sum(X * X, axis=0)

def sqr_norms_l2_rw(X):
    """
    Computes the l_2 norm of each row of a matrix
    """
    return jnp.sum(X * X, axis=1)


######################################
# Normalization of vectors
######################################

def normalize_l1(x):
    """Normalizes a vector by its l_1-norm
    """
    x = promote_arg_dtypes(x)
    x2 = jnp.abs(x)
    s = jnp.sum(x) + EPS
    return jnp.divide(x, s)

def normalize_l2(x):
    """Normalizes a vector by its l_2-norm
    """
    x = promote_arg_dtypes(x)
    s = jnp.sqrt(jnp.sum(x ** 2)) + EPS
    return jnp.divide(x, s)

def normalize_linf(x):
    """Normalizes a vector by its l_inf-norm
    """
    x = promote_arg_dtypes(x)
    s = jnp.max(jnp.abs(x)) + EPS
    return jnp.divide(x, s)

######################################
# Normalization of rows and columns
######################################


def normalize_l1_cw(X):
    """
    Normalize each column of X per l_1-norm
    """
    X = promote_arg_dtypes(X)
    X2 = jnp.abs(X)
    sums = jnp.sum(X2, axis=0) + EPS
    return jnp.divide(X, sums)

def normalize_l1_rw(X):
    """
    Normalize each row of X per l_1-norm
    """
    X = promote_arg_dtypes(X)
    X2 = jnp.abs(X)
    sums = jnp.sum(X2, axis=1) + EPS
    # row wise sum should be a column vector
    sums = jnp.expand_dims(sums, axis=-1)
    # now broadcasting works well
    return jnp.divide(X, sums)

def normalize_l2_cw(X):
    """
    Normalize each column of X per l_2-norm
    """
    X = promote_arg_dtypes(X)
    X2 = jnp.square(X)
    sums = jnp.sum(X2, axis=0) 
    sums = jnp.sqrt(sums)
    return jnp.divide(X, sums)

def normalize_l2_rw(X):
    """
    Normalize each row of X per l_2-norm
    """
    X = promote_arg_dtypes(X)
    X2 = jnp.square(X)
    sums = jnp.sum(X2, axis=1)
    sums = jnp.sqrt(sums)
    # row wise sum should be a column vector
    sums = jnp.expand_dims(sums, axis=-1)
    # now broadcasting works well
    return jnp.divide(X, sums)

